#include <ESP8266WiFi.h> // Importa a Biblioteca ESP8266WiFi
#include <PubSubClient.h> // Importa a Biblioteca PubSubClient
#include <DHT.h>             //Biblioteca do DHT

//defines:
//defines de id mqtt e tópicos para publicação e subscribe
// Tópicos
#define MQTTPUBTOPIC_T     "COPA/temp"   //Tópico para publicar leitura da temperatura
#define MQTTPUBTOPIC_H     "COPA/humi"   //Tópico para publicar leitura da umidade
#define MQTTPUBTOPIC_L     "COPA/lumi"   //Tópico para publicar leitura de luminosidade

#define ID_MQTT  "COPA"   //id mqtt (para identificação de sessão)
                           //IMPORTANTE: este deve ser único no broker (ou seja, 
                           //            se um client MQTT tentar entrar com o mesmo 
                           //            id de outro já conectado ao broker, o broker 
                           //            irá fechar a conexão de um deles).
                               

// WIFI
const char* SSID = "NOME_WIFI";         // SSID / nome da rede WI-FI que deseja se conectar
const char* PASSWORD = "SENHA_WIFI"; // Senha da rede WI-FI que deseja se conectar

// MQTT
const char* BROKER_MQTT = "URL_BROKER";   //URL do broker MQTT que se deseja utilizar
int BROKER_PORT = 11313;                         // Porta do Broker MQTT
const char* mqttUser = "USUARIO_MQTT";              //user
const char* mqttPassword = "SENHA_MQTT";     

// MQTT_mosquitto
const char* BROKER_MQTT_mosquitto = "URL_MOSQUITTO";   //URL do broker MQTT que se deseja utilizar
int BROKER_PORT_mosquitto = 1883;                         // Porta do Broker MQTT
const char* mqttUser_mosquitto = "";              //user
const char* mqttPassword_mosquitto = "";     

// Tipo do sensor DHT
#define DHTTYPE DHT22       //DHT22(AM2302/ AM2321)

// Configuração dos pinos do ESP8266 NodeMcu
#define PINDHT 4           //Pino para comunicação com DHT

// Inicialização das variáveis medidas
float tmp = 0.0;    // Temperatura
float hum = 0.0;    // Umidade
float lum = 0.0;    // Luminosidade

//Variáveis e objetos globais
WiFiClient espClient;             // Cria o objeto espClient
PubSubClient MQTT(espClient);     // Instancia o Cliente MQTT passando o objeto espClient
PubSubClient MQTT_mosquitto(espClient);     // Instancia o Cliente MQTT passando o objeto espClient
DHT dht(PINDHT, DHTTYPE);        //Instância do DHT

//Prototypes
void initSerial();
void initWiFi();
void initMQTT();
void initMQTT_mosquitto();
void reconectWiFi(); 
void mqtt_callback(char* topic, byte* payload, unsigned int length);
void mqtt_callback_mosquitto(char* topic, byte* payload, unsigned int length);
void VerificaConexoesWiFIEMQTT(void);
void InitOutput(void);

/* 
 *  Implementações das funções
*/
void setup() 
{
    //inicializações:
    InitOutput();
    initSerial();
    initWiFi();
    initMQTT();
    initMQTT_mosquitto();
}


void InitOutput(void)
{
    pinMode(2, OUTPUT);
    digitalWrite(2, HIGH);          
} 

void initSerial() 
{
    Serial.begin(115200);
}

void initWiFi() 
{
    delay(10);
    Serial.println("------Conexao WI-FI------");
    Serial.print("Conectando-se na rede: ");
    Serial.println(SSID);
    Serial.println("Aguarde");
    
    reconectWiFi();
}

void initMQTT() 
{
    MQTT.setServer(BROKER_MQTT, BROKER_PORT);   //informa qual broker e porta deve ser conectado
    MQTT.setCallback(mqtt_callback);            //atribui função de callback (função chamada quando qualquer informação de um dos tópicos subescritos chega)
}

void initMQTT_mosquitto() 
{
    MQTT_mosquitto.setServer(BROKER_MQTT_mosquitto, BROKER_PORT_mosquitto);   //informa qual broker e porta deve ser conectado
    MQTT_mosquitto.setCallback(mqtt_callback_mosquitto);            //atribui função de callback (função chamada quando qualquer informação de um dos tópicos subescritos chega)
}

void mqtt_callback(char* topic, byte* payload, unsigned int length) 
{
    String msg;

    //obtem a string do payload recebido
    for(int i = 0; i < length; i++) 
    {
       char c = (char)payload[i];
       msg += c;
    }
    Serial.print("Tópico Cloud: ");
    Serial.print(topic);
    Serial.print(" | ");
    Serial.println(msg);
}

void mqtt_callback_mosquitto(char* topic, byte* payload, unsigned int length) 
{
    String msg;

    //obtem a string do payload recebido
    for(int i = 0; i < length; i++) 
    {
       char c = (char)payload[i];
       msg += c;
    }
    Serial.print("Tópico Mosquitto: ");
    Serial.print(topic);
    Serial.print(" | ");
    Serial.println(msg);
}

void reconnectMQTT() 
{
    while (!MQTT.connected()) 
    {
        MQTT.connect("ESP8266", mqttUser, mqttPassword);
        Serial.print(" Tentando conectar ao Broker MQTT: ");
        Serial.println(BROKER_MQTT);
        if (MQTT.connect(ID_MQTT)) 
        {
            Serial.println("Conectado com sucesso ao broker MQTT!");
            MQTT.subscribe(MQTTPUBTOPIC_T); //Assinando o topico
            MQTT.subscribe(MQTTPUBTOPIC_L); //Assinando o topico
            MQTT.subscribe(MQTTPUBTOPIC_H); //Assinando o topico
        } 
        else 
        {
            Serial.println("Falha ao reconectar no broker.");
            Serial.println("Havera nova tentatica de conexao em 2s");
            delay(2000);
        }
    }
}

void reconnectMQTT_mosquitto() 
{
    while (!MQTT_mosquitto.connected()) 
    {
        MQTT_mosquitto.connect("ESP8266", mqttUser_mosquitto, mqttPassword_mosquitto);
        Serial.print(" Tentando conectar ao Broker MQTT Mosquitto: ");
        Serial.println(BROKER_MQTT_mosquitto);
        if (MQTT_mosquitto.connect(ID_MQTT)) 
        {
            Serial.println("Conectado com sucesso Mosquitto ao broker MQTT!");
            MQTT_mosquitto.subscribe(MQTTPUBTOPIC_T); //Assinando o topico
            MQTT_mosquitto.subscribe(MQTTPUBTOPIC_L); //Assinando o topico
            MQTT_mosquitto.subscribe(MQTTPUBTOPIC_H); //Assinando o topico
        } 
        else 
        {
            Serial.println("Falha ao reconectar no broker Mosquitto.");
            Serial.println("Havera nova tentatica de conexao em 2s");
            delay(2000);
        }
    }
}

void reconectWiFi() 
{
    //se já está conectado a rede WI-FI, nada é feito. 
    //Caso contrário, são efetuadas tentativas de conexão
    if (WiFi.status() == WL_CONNECTED)
        return;
        
    WiFi.begin(SSID, PASSWORD); // Conecta na rede WI-FI
    
    while (WiFi.status() != WL_CONNECTED) 
    {
        delay(100);
        Serial.print(".");
    }
  
    Serial.println();
    Serial.print("Conectado com sucesso na rede ");
    Serial.print(SSID);
    Serial.println("IP obtido: ");
    Serial.println(WiFi.localIP());
}


void VerificaConexoesWiFIEMQTT(void)
{
    if (!MQTT.connected()){ 
        reconnectMQTT();    //se não há conexão com o Broker, a conexão é refeita
        };
    if (!MQTT_mosquitto.connected()){
        reconnectMQTT_mosquitto(); 
    };
     reconectWiFi();        //se não há conexão com o WiFI, a conexão é refeita
}

void VerificaConexoesWiFIEMQTT_mosquitto(void)
{
    if (!MQTT_mosquitto.connected()){
        reconnectMQTT_mosquitto(); 
    };
     reconectWiFi();        //se não há conexão com o WiFI, a conexão é refeita
}



void readSensor2Mqtt()
{
  //Ler a temperatura ou humidade leva cerca de 250 milisegundos!
  float t = dht.readTemperature();
  float h = dht.readHumidity();
  float l = analogRead(A0);
  // Checando se qualquer leitura falhou
  if (isnan(h) || isnan(t))
  {
    for (int i=0; i<5; i++){
          t = dht.readTemperature();
          h = dht.readHumidity();
          if (t > 0 || h > 0 ){
                tmp = t;
                hum = h;
                //lum = l;
                  float ldr = l;
                  float b = 1023.0 / ldr;
                  float rLdr = 10000 * (b - 1);
                  float lux = 255.84 * pow(rLdr, -10/9) * 1000;
                  lum = lux;
                break;
           }
           else
           {
            //printSerial(true, "Falha na leitura do sensor!");
            tmp = 999;
            hum = 999; 
            //lum = l;
              float ldr = l;
              float b = 1023.0 / ldr;
              float rLdr = 10000 * (b - 1);
              float lux = 255.84 * pow(rLdr, -10/9) * 1000;
              lum = lux;
            }
      delay(2000);
      }
  }
  else{
    tmp = t;
    hum = h;
    //lum = l;
      float ldr = l;
      float b = 1023.0 / ldr;
      float rLdr = 10000 * (b - 1);
      float lux = 255.84 * pow(rLdr, -10/9) * 1000;
      lum = lux;
   
    delay(1000);
  }
}

void pubMqttBroker()
{
  //printSerial(true, "Fazendo a publicacao...");
  // Executando a leitura do sensor e atualizando variáveis tmp & hum
  String sTmp = String(tmp);
  String sHum = String(hum);
  String sLum = String(lum);
  
  char* msgMqtt_T = new char[sTmp.length() + 1];
  char* msgMqtt_H = new char[sHum.length() + 1];
  char* msgMqtt_L = new char[sLum.length() + 1];
  
  strcpy(msgMqtt_T, sTmp.c_str());
  strcpy(msgMqtt_H, sHum.c_str());
  strcpy(msgMqtt_L, sLum.c_str());

if (MQTT.connected()) {  
  MQTT.publish(MQTTPUBTOPIC_T, msgMqtt_T, true);
  MQTT.publish(MQTTPUBTOPIC_H, msgMqtt_H, true);
  MQTT.publish(MQTTPUBTOPIC_L, msgMqtt_L, true);
}

if (MQTT_mosquitto.connected()) {
  MQTT_mosquitto.publish(MQTTPUBTOPIC_T, msgMqtt_T, true);
  MQTT_mosquitto.publish(MQTTPUBTOPIC_H, msgMqtt_H, true);
  MQTT_mosquitto.publish(MQTTPUBTOPIC_L, msgMqtt_L, true);
} 
  delete [] msgMqtt_T;
  delete [] msgMqtt_H;
  delete [] msgMqtt_L;
}


//programa principal
void loop() 
{   
    readSensor2Mqtt();
    VerificaConexoesWiFIEMQTT();
    pubMqttBroker();
    MQTT.disconnect();
    VerificaConexoesWiFIEMQTT_mosquitto();
    pubMqttBroker();
    MQTT_mosquitto.disconnect();
    // Esta é a linha com o Sleep.
    ESP.deepSleep(10000000);//Dorme por 10 Segundos (Deep-Sleep em Micro segundos).
}
